#Created by Yitong Song in May, 2017
#coding:utf-8
#!/usr/bin/python

import json
import time
import urllib2
import urllib
import requests
import cookielib
import os.path
import commands
#import datetime
import logging

class AzkabanUtil:
    #*************************************************************************#
    #Constructor for Class AzkabanUtil                                        #
    #@params:                                                                 #
    #   - self: AzkabanUtil object                                            #
    #   - host: azkaban web serverhost                                        #
    #   - username: azkaban login azkaban_username                            #
    #   - password: azkaban login azkaban_password                            #
    #@return: None                                                            #
    #*************************************************************************#
    def __init__(self, host, username, password):
        self.azkaban_host = host
        self.azkaban_username = username
        self.azkaban_password = password


    #*************************************************************************#
    #Login to Azkaban web app and upload a project zip file                   #
    #more details on official documentation                                   #
    #http://azkaban.github.io/azkaban/docs/latest/#api-upload-a-project-zip   #
    #@params:                                                                 #
    #   - self: AzkabanUtil object                                            #
    #   - project_name: the name of azkaban project                           #
    #   - file: the zip file created to be uploaded, more requirements about  #
    #the zip file please see http://azkaban.github.io/docs/2.5/#upload-project#
    #@return: None                                                            #
    #*************************************************************************#
    def upload_a_project_zip(self, project_name, file):
        session_id = self.get_session_id(self.azkaban_username, 
        								 self.azkaban_password)
        url = "%s/manager"%(self.azkaban_host)
        command = '''curl -k -i -H "Content-Type: multipart/mixed" -X POST --form 'session.id=%s' --form 'ajax=upload' --form 'file=@%s;type=application/zip' --form 'project=%s' %s''' % (session_id,file,project_name,url)
        print '>>>>',command
        status,result = commands.getstatusoutput(command) #used for debugging
       	print ""
        return


    #*************************************************************************#
    #This API call schedules a flow by a cron Expression.                     #
    #more details on official documentation                                   #
    #http://azkaban.github.io/azkaban/docs/latest/#api-flexible-schedule      #
    #@params:                                                                 #
    #   - self: AzkabanUtil object                                            #
    #   - project_name: the name of the project.                              #
    #   - flow_name: the name of the flow   								  #
    #	- cron_expression: A CRON expression is a string comprising 6 or 7    #
    #fields separated by white space that represents a set of times.In azkaban#
    #we use Quartz Cron Format.                                               #
    #@return: None                                                            #
    #*************************************************************************#
    def schedule_flow(self, project_name, flow_name, cron_expression):
        session = self.get_session(project_name)
        url = "%s/schedule"%(self.azkaban_host)
        payload = {
            "ajax"                  : "scheduleCronFlow",
            "projectName"           : project_name,
            "flow"                  : flow_name,
            "cronExpression"        : cron_expression
        }        
        resp = session.post(url, data = payload)
        print(resp.text)
        return
        

    #*************************************************************************#
    #This API executes a flow via an ajax call, supporting a rich selection of#
    #different options. Running an individual job can also be achieved via    #
    #this API by disabling all other jobs in the same flow.                   #
    #More details on official documentation                                   #
    #http://azkaban.github.io/azkaban/docs/latest/#api-execute-a-flow         #
    #@params:                                                                 #
    #   - self: AzkabanUtil object                                            #
    #   - project_name: the name of the project.                              #
    #   - flow_name: the name of the flow   								  #
    #   - params: customized params, will replace the existed payload         #
    #@return: None                                                            #
    #*************************************************************************#
    def execute_a_flow(self, project_name, flow_name, params):
        session = self.get_session(project_name)
        url = "%s/executor"%(self.azkaban_host)
        payload = {
            "ajax"                  : "executeFlow",
            "project"               : project_name,
            "flow"                  : flow_name,
            "disabled"              : "[]",
            "failureEmailsOverride" : "false",
            "successEmailsOverride" : "false",
            "failureAction"         : "finishPossible",
            "notifyFailureFirst"    : "false",
            "notifyFailureLast"     : "false",
            "concurrentOption"      : "ignore"   
        }

        for k,v in params:
            payload[k] = v

        resp = session.post(url, data = payload)
        print(resp.text)
        return

    #*************************************************************************#
    #Given an execution id, this API call cancels a running flow. If the flow #
    #is not running, it will return an error message.                         #
    #More details on official documentation                                   #
    #http://azkaban.github.io/azkaban/docs/latest/#api-cancel-a-flow-execution#
    #@params:                                                                 #
    #   - self: AzkabanUtil object                                            #
    #   - project_name: the name of the project.                              #     #
    #   - execid: the execution id                                            #
    #@return: None                                                            #
    #*************************************************************************#
    def cancel_a_flow_execution(self, project_name, exec_id):
        session = self.get_session(project_name)
        url = "%s/executor"%(self.azkaban_host)
        payload = {
            "ajax"      : "cancelFlow",
            "execid"    : exec_id
        }
        resp = session.post(url, data = payload)
        print(resp.text)
        return


    #*************************************************************************#
    #Given an execution id, this API call fetches all the detailed information#
    #of that execution, including a list of all the job executions.           #
    #More details on official documentation                                   #
    #http://azkaban.github.io/azkaban/docs/latest/#api-fetch-a-flow-execution #
    #@params:                                                                 #
    #   - self: AzkabanUtil object                                            #
    #   - execid: the execution id                                            #
    #@return: None                                                            #
    #*************************************************************************#
    def fetch_a_flow_execution(self, project_name, exec_id):
        session = self.get_session(project_name)
        url = "%s/executor"%(self.azkaban_host)
        payload = {
            "ajax"      : "fetchexecflow",
            "execid"    : exec_id
        }
        resp = session.post(url, data = payload)
        print(resp.text)
        return


    #*************************************************************************#
    #Given a project name, and a certain flow, this API call provides a list  #
    #of corresponding executions. Those executions are sorted in descendent   #
    #submit time order. Also parameters are expected to specify the start     #
    #index and the length of the list. This is originally used to handle      #
    #pagination.															  #
    #More details on official documentation: http://azkaban.github.io/azkaban/#
    #docs/latest/#api-fetch-executions-of-a-flow  							  #
    #@params:                                                                 #
    #   - self: AzkabanUtil object                                            #
    #   - project_name: the name of the project.                              #
    #   - flow_name: the name of the flow                                     #
    # 	- num: the max length of the returned list		                      #
    #@return: None                                                            #
    #*************************************************************************#
    def fetch_executions_of_a_flow(self, project_name,flow_name, num): # latest
        session = self.get_session(project_name)
        url = "%s/manager"%(self.azkaban_host)
        payload = {
            "ajax"      : "fetchFlowExecutions",
            "flow"      : flow_name,
            "start"     : 0, 
            "length"    : num,
            "project"   : project_name,
        }
        resp = session.post(url, data = payload)
        print(resp.text)
        return


    #helper
    #*************************************************************************#
    #Build up a session and log in                                            #
    #@params:                                                                 #
    #   - self: AzkabanUtil object                                            #
    #   - username: azkaban login azkaban_username                            #
    #   - password: azkaban login azkaban_password                            #
    #@return:   															  #
    #	- the newly-created session id                                        #
    #*************************************************************************#
    def get_session_id(self, username, password):
        url = self.azkaban_host
        params = {}
        params['action'] = 'login'
        params['username']= username
        params['password']= password
        data = urllib.urlencode(params) 
        req = urllib2.Request(url,data)
        res_data = urllib2.urlopen(req)
        res = res_data.read()
        obj = json.loads(res)

        return  obj['session.id']

    #*************************************************************************#
    #Build up a session and log in                                            #
    #@params:                                                                 #
    #   - self: AzkabanUtil object                                            #
    #   - project_name: the name of the project.                              #
    #@return:                                                                 #
    #   - the newly-created session                                           #
    #*************************************************************************#
    def get_session(self, project_name):
        s = requests.Session()
        data = {
            'action' : 'login',
            'username' : self.azkaban_username,
            'password' : self.azkaban_password
        }
        url = self.azkaban_host
        s.get(url)
        res = s.post(url, data=data, json=None,)
        return s
        









#Created by Yitong Song in May, 2017
from job import * 
from azkaban_util import *
import job
import azkaban_util
import os
import tempfile
import os.path
import zipfile
import shutil

#*****************************************************************************#
#Constructor                                                                  #
#@params:                                                                     #
#   - self: the Project class project                                         #
#	- host: azkaban web server host                                           #
#@return: None                                                                #
#*****************************************************************************#
def __init__(self, host, username, password):
	self.azkaban_host = host
	self.azkaban_username = username
	self.azkaban_password = password


#*****************************************************************************#
#Merge the key and the value to a string in a certain form                    #
#@params:                                                                     #
#   - k: a key in a dict                                                      #
#	- v: a value in a dict                                                    #
#@return: 																	  #
#	- serialized: a string made up with k and v in a certain form             #
#*****************************************************************************#
def process_property(k,v):
		serialized = ""
		serialized = serialized + str(k) + "="
		if callable(v):  
			delimited = False
			for v_part in v:
				if delimited: 
					serialized = serialized + ","
				serialized += v_part
				delimited = True
		else:
			serialized += str(v)
		serialized += "\n"
		return serialized


#*****************************************************************************#
#Created the zip file and upload it.                                          #
#@params:                                                                     #
#   - project : the project needed to be handled                              #
#@return: None                                                                #
#*****************************************************************************#
def deploy(project):
	stg = tempfile.mkdtemp()
	to_zip = []
	print "staging job to" + str(stg)
	prop_file = stg + "/" + project.name + ".properties"
	f = open(prop_file, "w")

	for k in project.property.keys():
		if project.property[k] is None:
			f.write(process_property(k,''))
		else:
			f.write(process_property(k,project.property[k]))
	
	to_zip.append(prop_file)
	basename = os.path.basename(prop_file) #get the basename of the file
	print "created " + basename
	f.close()

	
	
	for job in project.jobs:
		job_file = stg + "/" + job.name + ".job"
		file = open(job_file, 'w')
		string = job.serialize()
		#print "string in the job_file:"
		#print string
		file.write(string)
		to_zip.append(job_file)
		basename = os.path.basename(job_file)
		print "created " + basename
		file.close()

	zip_file = stg + "/" 
	file_name = project.name + ".zip"
	zip_files(zip_file, to_zip, project.name)
	zip_file = zip_file + project.name + ".zip"
	print "created " + os.path.basename(zip_file)
	print zip_file

	

	azkUtil = AzkabanUtil(project.host, project.user, project.password)
	azkUtil.upload_a_project_zip(project.name, zip_file)
	print "uploaded to project %s" % project.name

	#shutil.rmtree(stg) #remove the temp file

	print "clean up"
	print "done!"
	return


#*****************************************************************************#
#Schedule a workflow using helper function                                    #
#@params:                                                                     #
#   - project : the project needed to be handled                              #
#@return: None                                                                #
#*****************************************************************************#
def schedule(project):
	for sch in project.schedules:
		project_name = sch['project_name']
		flow_name = sch['flow_name']
		if 'cron_expression' not in sch.keys():
			print "Error: no schedule in *.yaml"
			exit(0)
		cron_expression = sch['cron_expression']
		azkUtil = AzkabanUtil(project.host, project.user, project.password)
		azkUtil.schedule_flow(project_name, flow_name, cron_expression)
		
		
		
		
def execute(project, params):
	for sch in project.schedules:
		project_name = sch['project_name']
		flow_name = sch['flow_name']
		azkUtil = AzkabanUtil(project.host, project.user, project.password)
		azkUtil.execute_a_flow(project_name, flow_name, params)


def kill(project, exec_id):
	for sch in project.schedules:
		project_name = sch['project_name']
		flow_name = sch['flow_name']
		azkUtil = AzkabanUtil(project.host, project.user, project.password)
		azkUtil.cancel_a_flow_execution(project_name, exec_id)


def info(project, exec_id):
	for sch in project.schedules:
		project_name = sch['project_name']
		flow_name = sch['flow_name']
		azkUtil = AzkabanUtil(project.host, project.user, project.password)
		azkUtil.fetch_a_flow_execution(project_name, exec_id)


def fetch(project, exec_id, num):
	for sch in project.schedules:
		project_name = sch['project_name']
		flow_name = sch['flow_name']
		azkUtil = AzkabanUtil(project.host, project.user, project.password)
		azkUtil.fetch_executions_of_a_flow(project_name, flow_name, num)




#*****************************************************************************#
#Generate a zip file with given sources                                       #
#@params:                                                                     #
#   - dest: the path where the new zip file will be created                   #
#	- sources: the given files needed to be compressed                        #
#	- name: the name of the newly-created zip file                            #
#@return: None                                                                #
#*****************************************************************************#
def zip_files(dest, sources, name):
	name += ".zip"
	os.chdir(dest)
	zipFile = zipfile.ZipFile(name,'w',)
	for tz in sources:
		zipFile.write(tz)

	zipFile.close()



	


#Created by Yitong Song in May, 2017
from job import *
from project_deployer import *

#import job
#import project_deployer
import sys
import os.path

#*****************************************************************************#
#Check if files all exist 									                  #
#@params:                                                                     #
#   - jobs: the jobs in the Project object                                    #
#@return: 																	  #
#	- ok: whether the files all exist		              					  #
#*****************************************************************************#
def files_all_exist(jobs):
	ok = 1
	handle_missing_file = lambda file, name : sys.stderr.write("Missing " + file + " in job " + name + " .\n")
		
	for job in jobs:
		working_dir = job.options['working.dir']
		llist = ["command", "command.1", "command.2", "command.3", "command.4"]
		cmds = {}
		for cmd in llist:
			if cmd in job.options.keys():
				cmds[cmd] = job.options[cmd]
				fields = cmds[cmd].split(" ")
				
				if len(fields) < 2:
					break
				if fields[0] == 'python':
					have_python_file = 0
					for field in fields:
						if field.endswith('py'):
							have_python_file = 1
							break
					if not have_python_file:
						handle_missing_file("python file", job.name) 
						ok = 0

				elif fields[0] == 'ruby':
					have_ruby_file = 0
					for field in fields:
						if field.endswith('rb'):
							have_ruby_file = 1
							break
					if not have_ruby_file:
						handle_missing_file('ruby file', job.name)
						ok = 0

				for field in fields:
					if field.endswith('.py') or field.endswith('.rb') or field.endswith('.sql'):
						file = working_dir + "/" + field
						if os.path.isfile(file):
							handle_missing_file(file, job.name)
							ok = 0
	if ok:
		print "Files all exist"
	return ok
				

#*****************************************************************************#
#Check if denpendencies exist 								                  #
#@params:                                                                     #
#   - jobs: the jobs in the Project object                                    #
#@return: 																	  #
#	- ok: whether the dependencies all exist                                  #
#*****************************************************************************#
def dependencies_all_exist(jobs):
	ok = 1
	handle_missing_file = lambda job, missed_job : sys.stderr.write("Missing dept[" + missed_job + "] in job " + job + " .\n")
	job_names = []
	
	
	for j in jobs:
		job_names.append(j.name)
	
	for job in jobs:
		if job.options["dependencies"] is None:
			break
		depends = job.options[dependencies]
		if not depends:
			continue
		else:
			break
		if not isinstance(depends, list):
			depends = [depends]
		for missed in (depends - job_names):
			handle_missing_file(job.name, missed)
			ok = 0
	if ok:
		print "Dependencies all exist"
	return ok
	


#*****************************************************************************#
#The wrapper function checking the file existence and dependencies	          #
#@params:                                                                     #
#   - jobs: the jobs in the Project object                                    #
#@return: None																  #
#*****************************************************************************#	
def flow_valid(jobs):
	if files_all_exist(jobs) and dependencies_all_exist(jobs):
		print "OK"
	else: 
		print "Error"



#Created by Yitong Song in May, 2017
from project_deployer import *
#import project_deployer

class Job:
	# build the types of commands
	global types 
	cmd_types = ["command", "java", "javaprocess", "pig"]
	types = {}
	for cmd in cmd_types:
		types[cmd] = cmd


	#constructor
	def __init__(self, name, options):
		try:
			options
		except NameError:
			self.options = {}
		else:
			self.options = options

		if "type" not in self.options:
			self.options["type"] = types["command"]
		self.name = name
		self.types = types
		

	#*************************************************************************#
	#Merge the key and the value in a dict to a string in a certain form      #
	#@params:                                                                 #
	#   - self: a Job object                                                  #
	#@return: 																  #
	#	- serialized: a string made up with k and v in a certain form         #
	#*************************************************************************#
	def serialize(self):
		serialized = ""
		for k in self.options.keys():
			if self.options[k] is None:
				serialized += process_property(k,"")
			else:
				serialized += process_property(k,self.options[k])
		return serialized







#Created by Yitong Song in May, 2017
#!/usr/bin/python
# -*- coding: utf-8 -*-
import time
import yaml
import sys
import os
import job
from project_deployer import deploy
from azkaban_util import *
from project import *
from jobs_checker import *


class Project:
	#*************************************************************************#
    #Constructor for Project Class                                            #
    #@params:                                                                 #
    #   - self: the Project class project                                     #
    #	- file: the .yaml file with data needed                               #
    #@return: None                                                            #
    #*************************************************************************#
	def __init__(self, file):
		f = open(file)
		conf = yaml.load(f)
		self.host = conf['azkaban']['host']
		self.user = conf['azkaban']['user']
		self.password = conf['azkaban']['password']
		f.close()

	#getters
	def getName(self):
		return self.name

	def getProperty(self):
		return self.property

	def getHost(self):
		return self.host

	def getUser(self):
		return self.user

	def getPassword(self):
		return self.password

	def getJobs(self):
		return self.jobs

	def getSchedules(self):
		return self.schedules 

	#*************************************************************************#
    #Split jobs based on their azkaban flows.                                 #
    #The first job is always the flow's name (last job of the flow).          #
    #@params:                                                                 #
    #   - self: the Project class project                                     #
    #	- file: the .yaml file with data needed                               #
    #@return: None                                                            #
    #*************************************************************************#
	def load_jobs_from_file(self, file):
		f = open(file)
		conf = yaml.load(f)
		self.name = conf['project']['name']
		self.property = {"azkaban.user.home" : ""}
		if 'user_home' in conf['project']:
			self.property["azkaban.user.home"] = conf['project']['user_home']
		
		self.working_dir = conf['project']['azkaban_working_dir']
		self.job_working_dir = conf['project']['job_working_dir']
		self.user_to_proxy = conf['project']['user_to_proxy']
		self.jobs = []
		self.schedules = []
		

		for workflow in conf['workflows']:
			if 'schedule' not in workflow.keys():
				schedule_conf = {}
			else:
				schedule_conf = workflow['schedule']

			flow_name = None
			for job in workflow['jobs']:
				params = {
					"working.dir" : self.working_dir, 
					"user.to.proxy": self.user_to_proxy,
					"failure.emails": "" if schedule_conf == {} else schedule_conf['failure_emails'],
					"success.emails": "" if schedule_conf == {} else schedule_conf['success_emails'],
					'command' : "bash -c 'cd " + self.job_working_dir + " && " + job['command'] + "'",
					'dependencies' : None if job['dependencies'] == None else job['dependencies'],
				}
				flow_name = job['name']
				# initialize a new job and add it to the "jobs" list
				self.jobs.append(Job(job['name'], params)) 
			
			#initialize a new schedule coresponding to the new job and add it to the "schedules" list
			new_schedule = {
				'project_name' : conf['project']['name'],
				'flow_name' : flow_name
			}
			if 'schedule' in workflow.keys():
				new_schedule['cron_expression'] = schedule_conf['cron_expression']
			self.schedules.append(new_schedule)

		
		f.close()
		return

	
	#*************************************************************************#
    #Check whether the files in jobs are valid.                               #
    #@params:                                                                 #
    #   - home_dir: the home directory                                        #
    #@return: None                                                            #
    #*************************************************************************#
	def is_valid(self, home_dir):
		if home_dir:
			params = {'home_dir' : home_dir}
		flow_valid(self.jobs)
		return

def usage():
	print "Wrong Usage! Please see https://xiaohongshu.quip.com/HsImAChwIa8z for usage."

def main():
	index = len(sys.argv)
	if sys.argv[index-1] == "-c":
		print "Establishing new project"
		new_project = Project(sys.argv[1])
		new_project.load_jobs_from_file(sys.argv[1])
		print "Checking new project directory valid or not..."
		new_project.is_valid(os.environ["HOME"])
		exit(0)	

	file = ""
	for arg in sys.argv:
		if "yaml" in arg:
			file = arg


	project = Project(file)
	project.load_jobs_from_file(file)
	if sys.argv[1] == "upload":
		deploy(project)
	elif sys.argv[1] == "schedule":
		schedule(project)
	elif sys.argv[1] == "execute":
		execute(project, {})
	elif sys.argv[1] == "kill":
		exec_id = sys.argv[2]
		kill(project, exec_id)
	elif sys.argv[1] == "info":
		exec_id = sys.argv[2]
		info(project, exec_id)
	elif sys.argv[1] == "fetch":
		exec_id = sys.argv[2]
		num = sys.argv[3]
		fetch(project, exec_id, num)
	else:
		usage()

	
	

if __name__ ==  '__main__':
        main()



